#include <torch/extension.h>
#include <ATen/ATen.h>
#include <vector>

#define CHECK_CUDA(x) TORCH_CHECK(x.is_cuda(), #x " must be a CUDA tensor")
#define CHECK_CONTIGUOUS(x) TORCH_CHECK(x.is_contiguous(), #x " must be contiguous")
#define CHECK_INPUT(x) CHECK_CUDA(x); CHECK_CONTIGUOUS(x)

// Optimized block size for shared memory tiling
#define TILE_DIM 32

// Tiled matrix multiplication kernel using shared memory with loop unrolling and optimized block size
// Computes: C = A (MxD) * B (DxN), where A is x (reshaped) and B is clusters
__global__ void tiledMatMulOptimizedBlock(const float* __restrict__ A, const float* __restrict__ B, float* __restrict__ C,
                              int M, int D, int N) {
  // Each thread computes one element in the output matrix C
  // using tiling to reduce global memory accesses
  int row = blockIdx.y * TILE_DIM + threadIdx.y;
  int col = blockIdx.x * TILE_DIM + threadIdx.x;

  __shared__ float tile_A[TILE_DIM][TILE_DIM];
  __shared__ float tile_B[TILE_DIM][TILE_DIM];

  float sum = 0.0f;

  // Loop over tiles in the inner dimension
  for (int t = 0; t < (D + TILE_DIM - 1) / TILE_DIM; t++) {
    // Load element of A into shared memory if within bounds
    int a_col = t * TILE_DIM + threadIdx.x;
    if (row < M && a_col < D) {
      tile_A[threadIdx.y][threadIdx.x] = A[row * D + a_col];
    } else {
      tile_A[threadIdx.y][threadIdx.x] = 0.0f;
    }

    // Load element of B into shared memory if within bounds
    int b_row = t * TILE_DIM + threadIdx.y;
    if (b_row < D && col < N) {
      tile_B[threadIdx.y][threadIdx.x] = B[b_row * N + col];
    } else {
      tile_B[threadIdx.y][threadIdx.x] = 0.0f;
    }

    __syncthreads();

    // Multiply the two tiles together with loop unrolling
    #pragma unroll
    for (int i = 0; i < TILE_DIM; i++) {
      sum += tile_A[threadIdx.y][i] * tile_B[i][threadIdx.x];
    }
    __syncthreads();
  }

  // Write the result to global memory if within bounds
  if (row < M && col < N) {
    C[row * N + col] = sum;
  }
}

// Forward function for NetVLAD (no ghost clusters) using the shared memory optimized kernel
torch::Tensor forward(
    torch::Tensor x,                // B x N x D
    torch::Tensor clusters,         // D x (K+G)
    torch::Tensor clusters2,        // 1 x D x K
    torch::Tensor bn_weight,        // (K+G)
    torch::Tensor bn_bias,          // (K+G)
    torch::Tensor bn_running_mean,  // (K+G)
    torch::Tensor bn_running_var,   // (K+G)
    int64_t feature_size,           // D
    int64_t cluster_size,           // K
    bool is_training
) {
    CHECK_INPUT(x);
    CHECK_INPUT(clusters);
    CHECK_INPUT(clusters2);
    CHECK_INPUT(bn_weight);
    CHECK_INPUT(bn_bias);
    CHECK_INPUT(bn_running_mean);
    CHECK_INPUT(bn_running_var);

    int64_t B = x.size(0);
    int64_t N_points = x.size(1); // Number of descriptors/features per sample
    int64_t D = feature_size;
    int64_t K = cluster_size;
    int64_t KplusG = clusters.size(1); // Total clusters including ghost clusters

    // Reshape x: B x N x D -> (B*N) x D
    auto x_reshaped = x.reshape({B * N_points, D});

    // Ensure device consistency
    if (x_reshaped.device() != clusters.device()) {
        TORCH_CHECK(false, "x.device() != clusters.device()");
    }

    // Allocate memory for assignment: (B*N) x (K+G)
    auto assignment = torch::empty({B * N_points, KplusG}, x.options());

    // Launch tiled matrix multiplication kernel
    // x_reshaped: (B*N) x D, clusters: D x (K+G), output: (B*N) x (K+G)
    int M = B * N_points;     // Number of rows
    int inner = D;            // Shared inner dimension
    int N_dim = KplusG;       // Number of columns
    
    dim3 block(TILE_DIM, TILE_DIM);
    dim3 grid((N_dim + TILE_DIM - 1) / TILE_DIM, (M + TILE_DIM - 1) / TILE_DIM);

    tiledMatMulOptimizedBlock<<<grid, block>>>(x_reshaped.data_ptr<float>(), clusters.data_ptr<float>(), assignment.data_ptr<float>(), M, inner, N_dim);

    // Apply BatchNorm
    assignment = torch::batch_norm(
        assignment,
        bn_weight,
        bn_bias,
        bn_running_mean,
        bn_running_var,
        is_training,
        0.1,
        1e-5,
        true
    );

    // Apply Softmax along dim=1
    assignment = torch::softmax(assignment, 1); // (B*N) x (K+G)

    // Remove ghost assignments: keep first K columns
    assignment = assignment.narrow(1, 0, K); // (B*N) x K

    // Reshape assignment: (B*N) x K -> B x N x K
    assignment = assignment.reshape({B, N_points, K});

    // Compute a_sum = assignment.sum(dim=1, keepdim=true) // B x 1 x K
    auto a_sum = assignment.sum(1, /*keepdim=*/true);

    // Expand clusters2 to match batch size: clusters2 is originally 1 x D x K
    auto clusters2_exp = clusters2.expand({B, D, K}); // B x D x K

    // Compute a = a_sum * clusters2
    auto a = clusters2_exp * a_sum; // B x D x K

    // Transpose assignment: B x N x K -> B x K x N
    assignment = assignment.transpose(1, 2);

    // Reshape x back to B x N x D
    auto x_orig = x.reshape({B, N_points, D});

    // Compute vlad = assignment @ x_orig; assignment: B x K x N, x_orig: B x N x D
    auto vlad = torch::bmm(assignment, x_orig); // B x K x D

    // Transpose vlad to B x D x K
    vlad = vlad.transpose(1, 2);

    // Compute vlad = vlad - a
    vlad = vlad - a;

    // L2 intra-normalization along feature dimension D
    vlad = torch::nn::functional::normalize(
        vlad, torch::nn::functional::NormalizeFuncOptions().p(2).dim(1));

    // Flatten vlad: B x D x K -> B x (D*K)
    vlad = vlad.reshape({B, D * K});

    // L2 normalization along the flattened dimension
    vlad = torch::nn::functional::normalize(
        vlad, torch::nn::functional::NormalizeFuncOptions().p(2).dim(1));

    return vlad;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "NetVLAD forward with optimized block size and loop unrolling");
}
